/* This programs creates eight output files in ASCII format.
	Samples are read in from a binary file.  */



/* This program calculates the DFT at a small
	number of frequencies using Goertzel's Algorithm  */

 #include <stdio.h>
 #include <stdlib.h>
 #include <math.h>

#define TRUE 1
#define FALSE 0
#define SIGNAL_CUTOFF 900
#define N 1148
#define FS 8000.0
#define PI 3.14159265359
#define F0 697.0
#define F1 770.0
#define F2 852.0
#define F3 941.0
#define F4 1209.0
#define F5 1336.0
#define F6 1477.0
#define F7 1633.0
#define SCALE_FACTOR 25000.0

struct freq
{ double k;
  int n;
};

float xr[N],xfr[N],xfi[N],c[8],s[8],w[N];
struct freq k[9];
int index,nmax;
unsigned int wsize;
short int sample[N];
char weight;
unsigned long loop_count;

void goertzel(short int sample[])
{ float vr1,vr2,temp;
  int i,j,x;

  /*

  k[j].k determines the frequency at which the DFT is computed.
  k[j].n determines the number of data samples to use.

  */
  /* add next w samples to circular buffer */
  for(i=0;i<wsize;i++)
  {
	 xr[index] = (float)(sample[i])/SCALE_FACTOR;
	 index++;
	 if(index >= nmax)
	 { index = 0; }
  }

	for(j=0;j<8;j++)
	{
	 vr1 = 0;
	 vr2 = 0;
	 i=index+(nmax-k[j].n);
	 if(i>=nmax)
		i= i-nmax;
 /* peel off first loop for circular pointer adjustment  */
	 x=0;
	 temp = vr1;
	 vr1 = (2.0*c[j]*vr1) - vr2 - w[x]*xr[i];
	 vr2 = temp;
	 x++;
	 i++;
	 if(i>=nmax) i=0;
	 while(i!=index)
	 {
		temp = vr1;
		vr1 = (2.0*c[j]*vr1) - vr2 - w[x]*xr[i];
		vr2 = temp;
		x++;
		i++;
		/* if(i != index) i++; */
		if(i>=nmax) i=0;
	 }
	 xfr[j] = (c[j]*vr1) - vr2;
	 xfi[j] = s[j]*vr1;
	}
	return;
}

main(int argc, char *argv[])
{  char sourcefile[20],targetfile[8][20];
	FILE *fptr1, *fptr[8];
	double magnitude;
	short int mag_out;
	int i, j, n, done, count,temp;
	float f[9],wn,phi,t;
	float unit_factor, m;
	int ascii,optimize;

	if(argc < 4)
	{
	  printf("\n\nUSAGE: GOERTZ5B filespec framesize windowsize [-O] [-A] [-T] [-Mx] [-LD]\n\n");
	  exit(-1);
	}

	strcpy(sourcefile,argv[1]);
	n = atoi(argv[2]);
	wsize = atoi(argv[3]);
	t = n/FS;

	optimize = 0;
	ascii = 0;
	weight = 0;
	i=4;
	while(i<argc)
	{
	  if(strcmp("-O",argv[i]) == 0) optimize = 1;
	  else if(strcmp("-A",argv[i]) == 0) ascii = 1;
	  else if(strcmp("-T",argv[i]) == 0) weight = 'T';
	  else if(strcmp("-V",argv[i]) == 0) weight = 'V';
	  else if(strcmp("-H",argv[i]) == 0) weight = 'H';
	  else if((strstr("-M",argv[i]) == 0) && (strcmp("-M",argv[i]) == 0 ))
	  {
		 m = (float)atof(argv[i]+2);
		 if(m!=0)
		  weight = 'M';
		 else
		 {
			printf("\n\nUSAGE: GOERTZ5B filespec framesize windowsize [-O] [-A] [-T] [-Mx] [-LD]\n\n");
			exit(-1);

		 }
	  }
	  else if(strcmp("-LD",argv[i]) == 0) weight = 'D';
	  i++;
	}


	fptr1 = fopen(sourcefile,"rb");
	if(fptr1 == NULL)
	{ printf("\rCannot open input file %s.   Program terminating.\n\n",sourcefile);
	  exit(-1);
	}

	for(i=0;i<8;i++)
	{
	  j=0;
	  while((sourcefile[j] != 0) && (sourcefile[j] != '.'))
	  { targetfile[i][j] = sourcefile[j];
		 j++;
	  }
	  targetfile[i][j] = '.';
	  targetfile[i][j+1] = 0x30 + i;
	  targetfile[i][j+2] = 0;

	  if(ascii == 0) fptr[i] = fopen(targetfile[i],"wb");
	  else fptr[i] = fopen(targetfile[i],"w");
	  if(fptr[i] == NULL)
	  {
		 printf("\n\nERROR: Cannot open output file %s.\n\n",targetfile[i]);
		 exit(-1);
	  }
	}


  /* initialize the frequencies of interest */
  nmax = 0;

  f[0] = F0;
  f[1] = F1;
  f[2] = F2;
  f[3] = F3;
  f[4] = F4;
  f[5] = F5;
  f[6] = F6;
  f[7] = F7;
  f[8] = 0;

  nmax = 0;
  i=0;
  while(i<8)
  {
	 k[i].n = n;
	 /*
	 while(fabs((double)(FS*(floor((double)(f[i]*k[i].n/FS))/k[i].n) - f[i])) > 0.8)
		k[i].n--;
	 */
	 if(optimize == 1)
	 {
		j=n;
		while(j > (n-100))
		{
		  if(fabs((double)(FS*(floor((double)(f[i]*j/FS))/(double)j) - f[i])) < fabs((double)(FS*(floor((double)(f[i]*k[i].n/FS))/k[i].n) - f[i])) )
			 k[i].n = j;
		  j--;
		}
	 }
	 if(k[i].n > nmax) nmax = k[i].n;
	 k[i].k = (float)((f[i]*(float)k[i].n)/FS);
	 printf("\nk[%i].n = %i      k[%i].k = %f     f = %4.0f    df = %lf",i,k[i].n,i,k[i].k,(float)(FS*k[i].k/k[i].n),fabs((double)(FS*(double)k[i].k/(double)k[i].n) - f[i]));
	 i++;

  }
  k[i].n = 0;
  k[i].k = 0;

  for(i=0;i<N;i++)
  { xr[i] = 0;
    xfr[i] = 0;
	 xfi[i] = 0;
  }

  /* initialize the sin and cos tables */
  for(i=0;i<8;i++)
  {
	 wn=6.2831853/((float)(k[i].n));
	 phi=wn*((float)k[i].k);
	 c[i] = (float)cos((double)phi);
	 s[i] = (float)sin((double)phi);
  }


	/* compute sample weightings for each sample */
	 switch(weight)
	 {
	  case 'D':            /* linearly decreasing window  */
		 unit_factor = 0;
		 for(i=0;i<nmax;i++)
		 {
			w[i] = ((float)(nmax)-(float)i)/(float)(nmax);
			unit_factor += w[i];
		 }
		 break;
	  case 'T':           /* triangular window  */
		 unit_factor = 0;
		 for(i=0;i<nmax;i++)
		 {
			if(i<(nmax/2)) w[i] = 2.0*(float)i/(float)(nmax-1);
			else
			{
			  w[i] = -2.0*(float)i/(float)(nmax-1)+2.0;
			}
			unit_factor += w[i];
		 }
		 break;
	  case 'V':          /*  Von Hann (raised cosine) window  */
		 unit_factor = 0;
		 for(i=0;i<nmax;i++)
		 {
			w[i] = 0.5 + 0.5*(float)cos(PI*(double)i/(double)nmax);
			unit_factor += w[i];
		 }
		 break;
	  case 'H':
		 unit_factor = 0;
		 for(i=0;i<nmax;i++)
		 {
			w[i] = 0.54 - 0.46*(float)cos(2.0*PI*(double)i/(double)nmax);
			unit_factor += w[i];
		 }
		 break;
	  case 'M':       /* modified Vonn Hann Window    */
		 unit_factor = 0;
		 for(i=0;i<nmax;i++)
		 {
			w[i] = pow(((1.0 - cos(2.0*PI*(double)i/(double)nmax))/2.0),(double)m);
			unit_factor += w[i];
		 }
		 break;
	  default:              /* rectangular window   */
		 unit_factor = 0;
		 for(i=0;i<nmax;i++)
		 {
			w[i] = 1.0;
			unit_factor += w[i];
		 }
		 break;
	 }

  index = 0;
  printf("\nWorking.");
  done = FALSE;
  while(done == FALSE)
  { temp++;
	 i=0;
	 while((i<wsize)&&(done == FALSE))
	 {
		count = fread( &sample[i], (size_t)2,1,fptr1);
		if(count != 1)
		{
		  done = TRUE;
		}
		i++;
	 }
	 if(done == FALSE)
	 {
		goertzel(sample);
		for(i=0;i<8;i++)
		{
		  temp = sqrt((double)(xfr[i]*xfr[i]+xfi[i]*xfi[i]))/unit_factor;
		  if(temp<0)
		    {
		      printf("temp = %f\n",temp);
		      exit(0);
		    }
		  magnitude = 10.0*log10(sqrt((double)(xfr[i]*xfr[i]+xfi[i]*xfi[i]))/unit_factor);
		  mag_out = (short int)magnitude;
		  if(ascii == 1) count = fprintf(fptr[i],"%lf\n",magnitude);
		  else count = fwrite(&mag_out, (size_t)2,1,fptr[i]);
		  if(count < 1)
		  {
			 printf("\n\nERROR writing to output file %s\n\n",targetfile[i]);
			 fclose(fptr1);
			 for(j=0;j<8;j++)
			 {
				fclose(fptr[j]);
			 }
			 exit(-1);
		  }

		}
	 }
	 loop_count++;
	 if(fmod((double)loop_count,(double)(10.0)) == 0.0) printf(".");

  }
	printf("\n\nDone....\n");
	fclose(fptr1);
	for(i=0;i<8;i++)
	{
	  fclose(fptr[i]);
	}
	exit(0);
}
