
/* This program calculates the DFT at a small
   number of frequencies using Goertzel's Algorithm  */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#define TRUE 1
#define FALSE 0
#define PI 3.14159265359
#define F0 697.0
#define F1 770.0
#define F2 852.0
#define F3 941.0
#define F4 1209.0
#define F5 1336.0
#define F6 1477.0
#define F7 1633.0
#define SCALE_FACTOR 10000.0
#define SAMPLES_TO_READ 1024
#define HIT_MISS_DIFF  4.0
#define MINSD 40        /* minimum signal duration in ms */
#define MINID 40        /* minimum interdigit time in ms */
#define MINCYC 93       /* minimum cycle time in ms */
#define MAX_TWIST 12.0
#define MIN_SIG_HIT -25.0
#define START_LOOP 0
#define FREQ_CHANGE 1.01
char keys[16],which_key[4][8],last_key,last_key_returned;
float xr[SAMPLES_TO_READ],xfr[8],xfi[8],mag[8],c[8],s[8],w[SAMPLES_TO_READ];
float k[8],unit_factor;
unsigned int fcurr,icurr,fprev,iprev;
int n,index,head,tail,twist,level_check_low,level_check_high;
unsigned int wsize,fs;
short int sample[SAMPLES_TO_READ];
char weight;
unsigned long loop_count,tempi;
float sigdur;

int timing ()
{   int i;
   
    
    
    i=0;
    while(keys[i] == 0) i++;
    if(i<16)   /* a key is currently pressed  */
      {
	fcurr = loop_count;
	if(keys[i] != last_key)
	  {
	    if(last_key != 0) fprev = loop_count;
	    icurr = loop_count;
	    last_key = keys[i];
	  }
      }
    else   /* no key pressed */
      {
	if(last_key != 0)
	  {
	    fprev = loop_count-icurr;
	    last_key = 0;
	    last_key_returned = 0;
	  }
	iprev = 0;
	loop_count = 0;
	icurr = 0;
	fcurr = 0;
      }
    
    /* calculate signal duration, interdigit time, cycle time */
    sigdur = (float)((float)(fcurr - icurr)*wsize*1000.0/fs);
    /*************************************************************
      
      ---we don't really care how short these are---
      
      indig = (int)((fprev - icurr)*wsize/fs);
      cycle = (int)((icurr - iprev)*wsize/fs);
      
      *************************************************************/
    
    /*****************************************************************
      
      return TRUE iff sigdur is O.K. and if this is a new keypress
      
      *******************************************************************/
    
    if ((sigdur>=MINSD) && (keys[i] != last_key_returned))
      {
	last_key_returned = keys[i];
	return(TRUE);
      }
    else
      {
	return(FALSE);
      }
  }

void level_check()
{
  float max_low=-100,max_high=-100;
  int i,hit_low,hit_high;
  
  /*Find the highest values for the low and high frequencies*/
  /*and where they occur.*/
  for(i=0;i<4;i++)
    {
      if(max_low<mag[i])
	{
	  max_low=mag[i];
	  hit_low=i;
	}
      if(max_high<mag[i+4]) 
	{
	  max_high=mag[i+4];
	  hit_high=(i+4);
	}
    }
  
  /*Check for proper twist*/
  if(fabs((double)(max_high-max_low))<MAX_TWIST)
    twist=1;
  else
    twist=0;
  
  /*Check maximum levels vs. other levels to determine whether or*/
  /*not our max levels are hits*/
  level_check_low=0;
  level_check_high=0;
  for(i=0;i<4;i++)
    {
      if((hit_low!=i)&&((max_low-mag[i])>HIT_MISS_DIFF))
	level_check_low+=1;
      if((hit_high!=(i+4))&&((max_high-mag[i+4])>HIT_MISS_DIFF))
	level_check_high+=1;
    }
  /*If all parameters are met, update keys array properly*/
  if((twist==1)&&(level_check_low==3)&&(level_check_high==3)&&(max_high > MIN_SIG_HIT)&&(max_low > MIN_SIG_HIT))
    keys[which_key[hit_low][hit_high]]=which_key[hit_low][hit_high-4];
  
}

void goertzel()
{ float vr1,vr2,temp;
  int i,j,x;
  
  for(j=0;j<8;j++)
    {
      vr1 = 0;
      vr2 = 0;
      i=tail+1;
      if(i>=SAMPLES_TO_READ) i=0;
      /* peel off first loop for circular pointer adjustment  */
      x=0;
      temp = vr1;
      vr1 = (2.0*c[j]*vr1) - vr2 - w[x]*xr[i];
      vr2 = temp;
      x++;
      i++;
      if(i>=SAMPLES_TO_READ) i=0;
      while(x<n)
	{
	  temp = vr1;
	  vr1 = (2.0*c[j]*vr1) - vr2 - w[x]*xr[i];
	  vr2 = temp;
	  x++;
	  i++;
	  if(i>=SAMPLES_TO_READ) i=0;
	}
      xfr[j] = (c[j]*vr1) - vr2;
      xfi[j] = s[j]*vr1;
      temp = (float)sqrt((double)(xfr[j]*xfr[j]+xfi[j]*xfi[j]));
      if(temp > 0) mag[j] = 10.0*log10((double)(temp/unit_factor));
      else mag[j] = -150.0;
    }
  return;
}

main(int argc, char *argv[])
{  char sourcefile[80];
   FILE *fptr;
   int i, j, done, count , diff, x, old_timer=FALSE, last_i;
   float f[9],wn,phi;
   float m,delta;
   

   
   if(argc < 2)
     {
       printf("\n\nUSAGE: DTMF.EXE framesize windowsize sample_freq [-T] [-M d m] [-LD]\n\n");
       exit(-1);
     }
   

   fs = atoi(argv[1]);
   n=330;
   wsize=80;
   weight='H';
   
   last_key_returned = 0;
   last_key = 0;
   i=4;
   tempi = 0;
   fptr = stdin;
   if(fptr == NULL)
     { printf("\rCannot open %s.   Program terminating.\n\n",sourcefile);
       exit(-1);
     }
   
   printf("\nThe DTMF Guys - DTMF Detector v1.0\n");
   printf("DTMF Detection Group - Mississippi State University\n");
   printf("ECE Deptartment DSP Conference Fall 1995\n");
   printf("\nWaiting for keypresses\n");
   /* initialize the frequencies of interest */
   
   f[0] = F0;
   f[1] = F1;
   f[2] = F2;
   f[3] = F3;
   f[4] = F4;
   f[5] = F5;
   f[6] = F6;
   f[7] = F7;
   f[8] = 0;
   
   for(i=0;i<8;i++) k[i] = (float)((f[i]*(float)n)/fs);
   
   for(i=0;i<8;i++)
     {
       xfr[i] = 0;
       xfi[i] = 0;
     }
   
   for(i=0;i<SAMPLES_TO_READ;i++)
     {
       xr[i] = 0;
     }
   
   /* initialize the sin and cos tables */
   for(i=0;i<8;i++)
     {
       wn=6.2831853/((float)(n));
       phi=wn*k[i];
       c[i] = (float)cos((double)phi);
       s[i] = (float)sin((double)phi);
     }
   
   for(i=0;i<4;i++)
     {
       for(j=4;j<8;j++)
	 {
	   switch(i)
	     {
	     case 0:
	       switch(j)
		 {
		 case 4:
		   which_key[i][j]=0;
		   which_key[i][j-4]='1';
		   break;
		 case 5:
		   which_key[i][j]=1;
		   which_key[i][j-4]='2';
		   break;
		 case 6:
		   which_key[i][j]=2;
		   which_key[i][j-4]='3';
		   break;
		 case 7:
		   which_key[i][j]=3;
		   which_key[i][j-4]='A';
		   break;
		 }
	       break;
	     case 1:
	       switch(j)
		 {
		 case 4:
		   which_key[i][j]=4;
		   which_key[i][j-4]='4';
		   break;
		 case 5:
		   which_key[i][j]=5;
		   which_key[i][j-4]='5';
		   break;
		 case 6:
		   which_key[i][j]=6;
		   which_key[i][j-4]='6';
		   break;
		 case 7:
		   which_key[i][j]=7;
		   which_key[i][j-4]='B';
		   break;
		 }
	       break;
	     case 2:
	       switch(j)
		 {
		 case 4:
		   which_key[i][j]=8;
		   which_key[i][j-4]='7';
		   break;
		 case 5:
		   which_key[i][j]=9;
		   which_key[i][j-4]='8';
		   break;
		 case 6:
		   which_key[i][j]=10;
		   which_key[i][j-4]='9';
		   break;
		 case 7:
		   which_key[i][j]=11;
		   which_key[i][j-4]='C';
		   break;
		 }
	       break;
	     case 3:
	       switch(j)
		 {
		 case 4:
		   which_key[i][j]=12;
		   which_key[i][j-4]='*';
		   break;
		 case 5:
		   which_key[i][j]=13;
		   which_key[i][j-4]='0';
		   break;
		 case 6:
		   which_key[i][j]=14;
		   which_key[i][j-4]='#';
		   break;
		 case 7:
		   which_key[i][j]=15;
		   which_key[i][j-4]='D';
		   break;
		 }
	       break;
	     }
	 }
     }
   
   for(i=0;i<16;i++) keys[i]=0;
   
   
   /* compute sample weightings for each sample */
   switch(weight)
     {
     case 'D':            /* linearly decreasing window  */
       unit_factor = 0;
       for(i=0;i<n;i++)
	 {
	   w[i] = ((float)(n)-(float)i)/(float)(n);
	   unit_factor += w[i];
	 }
       break;
     case 'T':           /* triangular window  */
       unit_factor = 0;
       for(i=0;i<n;i++)
	 {
	   if(i<(n/2)) w[i] = 2.0*(float)i/(float)(n-1);
	   else
	     {
	       w[i] = -2.0*(float)i/(float)(n-1)+2.0;
	     }
	   unit_factor += w[i];
	 }
       break;
     case 'N':           /* Hamming with different frequency.  */
       unit_factor = 0;
       for(i=0;i<n;i++)
	 {
	   w[i] = 0.54 - 0.46*(float)cos(2.0*PI*(double)i*FREQ_CHANGE/(double)n);
	   unit_factor += w[i];
	 }
       break;
     case 'V':          /*  Von Hann (raised cosine) window  */
       unit_factor = 0;
       for(i=0;i<n;i++)
	 {
	   w[i] = 0.5 + 0.5*(float)cos(PI*(double)i/(double)n);
	   unit_factor += w[i];
	 }
       break;
     case 'H':
       unit_factor = 0;
       for(i=0;i<n;i++)
	 {
	   w[i] = 0.54 - 0.46*(float)cos(2.0*PI*(double)i/(double)n);
	   unit_factor += w[i];
	 }
       break;
     case 'M':       /* modified Vonn Hann Window    */
       unit_factor = 0;
       for(i=0;i<n;i++)
	 {
	   w[i] = pow(0.5+delta - (0.5-delta)*cos(2.0*PI*(double)i/(double)n),(double)m);
	   unit_factor += w[i];
	 }
       break;
     default:              /* rectangular window   */
       unit_factor = 0;
       for(i=0;i<n;i++)
	 {
	   w[i] = 1.0;
	   unit_factor += w[i];
	 }
       break;
     }
   x=1;
   fcurr = 0; icurr = 0; fprev = 0; iprev = 0;
   index = 0;
   head = 0;
   tail = 0;
   loop_count = 0;
   done = FALSE;
   while(done == FALSE)
     {
       /* add next wsize samples to circular buffer */
       if((head-tail)>=0) diff = head - tail;
       else diff = SAMPLES_TO_READ + head - tail;
       while((diff<n) && (done == FALSE))
	 {
	   count = fread( sample, (size_t)2,SAMPLES_TO_READ-diff-1,fptr);
	   i=0;
	   while(i < count)
	     {
	       head++;
	       if(head>=SAMPLES_TO_READ) head = 0;
	       xr[head] = (float)sample[i]/SCALE_FACTOR;
	       i++;
	     }
	   if((head-tail)>=0) diff = head - tail;
	   else diff = SAMPLES_TO_READ + head - tail;
	   if((feof(fptr) != 0) || (ferror(fptr) != 0)) done = TRUE;
	 }
       
       while(diff >= n)
	 {
	   if(tempi > START_LOOP)
	     goertzel();
	   level_check();
	   if(timing() == TRUE)
	     {
	       i=0;
	       old_timer=TRUE;
	       while(keys[i] == 0) i++;
	       printf("\rDigit %c        Start Time: %f",keys[i],(float)(wsize*tempi+n)/(float)fs);
	       last_i=i;
	     }
	   if((old_timer==TRUE)&&(keys[last_i]==0))
	     {
	     old_timer=FALSE;
	     printf("       Stop Time: %f\n",(float)((float)(wsize*tempi+n)/(float)fs));
	   }
	   for(i=0;i<16;i++) keys[i]=0;
	   tail += wsize;
	   if(tail >= SAMPLES_TO_READ) tail -= SAMPLES_TO_READ;
	   if((head-tail)>=0) diff = head - tail;
	   else diff = SAMPLES_TO_READ + head - tail;
	   loop_count++;
	   tempi++;
	   if(((float)fmod((double)(tempi),(double)(20.0))==0)&&(old_timer==FALSE))
             {
	       if(x==1){printf("\r\\"); x++;}
              else if(x==2){printf("\r|"); x++;}
              else if(x==3){printf("\r/"); x++;}
              else if(x==4){printf("\r-"); x=1;}
	    }
	   fflush(stdout);
	 }
     }
   printf("\nFinal time is %f\n\n",(float)((float)(wsize*tempi+n)/(float)fs));
   fclose(fptr);
   exit(0);
 }



































